# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass
from time import time
from typing import Dict, List, Optional, Tuple

import torch

from atek.evaluation.static_object_detection.eval_obb3_metrics_utils import (
    bb3_valid,
    box3d_overlap_wrapper,
    box3d_volume,
)

from torchmetrics.detection.mean_ap import (
    _fix_empty_tensors,
    BaseMetricResults,
    MARMetricResults,
    MeanAveragePrecision,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MAPMetricResults3D(BaseMetricResults):
    """Class to wrap the final mAP results."""

    __slots__ = (
        "map",
        "map_25",
        "map_50",
        "map_small",
        "map_medium",
        "map_large",
    )


class AtekObb3Metrics(torch.nn.Module):
    """
    Metrics class that works with ATEK obb3 gt dict format.
    This is a thin wrapper around MeanAveragePrecision3D metrics class, to be able to use it with ATEK obb3 gt dict format.
    It is a torch.nn.Module to be able to behave like a torchmetrics object.

    This class is easily extensible to support more input format, by adding / overloading the `update()` API function.
    """

    def __init__(
        self,
        class_metrics: bool = False,
        global_name_to_id: Optional[Dict] = None,
        ret_all_prec_rec: Optional[bool] = False,
        max_detection_thresholds: Optional[List[float]] = None,
    ) -> None:
        """
        Args:
            cam_ids (list): list of camera ids to evaluate
            cam_names (list): list of camera names to evaluate
            class_metrics (bool): if True, computes per-class metrics
            max_detection_thresholds (list): list of maximum number of detections to evaluate
        """

        super().__init__()
        if max_detection_thresholds is None:
            # max number of detections to evaluate - 220 is sufficient for ASE scenes
            max_detection_thresholds = [220]

        self.metric_3d = MeanAveragePrecision3D(
            class_metrics=class_metrics,
            max_detection_thresholds=max_detection_thresholds,
            ret_all_prec_rec=ret_all_prec_rec,
        )
        self.sem_id_to_name = None

        if global_name_to_id is not None:
            self.sem_id_to_name = {
                int(sem_id): name for name, sem_id in global_name_to_id.items()
            }

    def update(self, prediction: Dict, target: Dict):
        """
        Update from prediction and target, in ATEK obb3 gt dict format.
        """
        # Compute bbox corners in world frame, as tensors of shape [N, 8, 3]
        pred_bb3corners_in_world = prediction["bbox_corners_in_world"]
        target_bb3corners_in_world = target["bbox_corners_in_world"]

        # convert to lists
        pred_list = [
            {
                "boxes": pred_bb3corners_in_world,
                "scores": prediction["confidence_scores"],
                "labels": prediction["category_ids"],
            }
        ]
        target_list = [
            {
                "boxes": target_bb3corners_in_world,
                "labels": target["category_ids"],
            }
        ]
        self.metric_3d.update(pred_list, target_list)

    def forward(self, prediction: Dict, target: Dict):
        self.update(prediction, target)
        return self.compute()

    def compute(self):
        # only compute the 3D metrics (mAP 3D)
        t0 = time()

        metrics = {}
        logger.info(f"Computing 3D obb metric")
        metrics_3d_result = self.metric_3d.compute(self.sem_id_to_name)

        t1 = time()
        logger.info(f"DONE Computing 3D obb metric in {t1-t0} seconds")

        for metric_name, val in metrics_3d_result.items():
            metrics[f"{metric_name}_3D"] = val
        return metrics

    def reset(self):
        for metric in self.metric_3d.values():
            metric.reset()


class MeanAveragePrecision3D(MeanAveragePrecision):
    def __init__(
        self,
        max_detection_thresholds: Optional[List[int]] = None,
        class_metrics: bool = False,  # compute per class metrics
        ret_all_prec_rec: bool = False,  # return all precision and recall values
    ) -> None:  # type: ignore
        # TODO: check for torchvision version

        # Set default IOU and recall thresholds
        iou_thresholds = torch.linspace(
            0.05, 0.5, round((0.5 - 0.05) / 0.05) + 1, dtype=torch.float64
        ).tolist()
        rec_thresholds = torch.linspace(
            0.0, 1.00, round(1.00 / 0.01) + 1, dtype=torch.float64
        ).tolist()
        super().__init__(
            iou_thresholds=iou_thresholds,
            rec_thresholds=rec_thresholds,
        )

        max_det_thr, _ = torch.sort(
            torch.IntTensor(max_detection_thresholds or [1, 10, 100])
        )
        self.max_detection_thresholds = max_det_thr.tolist()

        self.class_metrics = class_metrics

        # important to overwrite after the __init__() call since they are otherwise overwritten by super().__init__()
        self.bbox_area_ranges = {"all": (0, 1e5)}
        self.ret_all_prec_rec = ret_all_prec_rec
        self.eval_imgs = [] if self.ret_all_prec_rec else None

    def update(self, preds: List[Dict[str, torch.Tensor]], target: List[Dict[str, torch.Tensor]]) -> None:  # type: ignore
        """Add detections and ground truth to the metric.

        Args:
            preds: A list consisting of dictionaries each containing the key-values
            (each dictionary corresponds to a single image):
            - ``boxes``: ``torch.FloatTensor`` of shape
                [num_boxes, 8, 3] containing `num_boxes` detection boxes of the format
                specified in the contructor. By default, this method expects
                (4) +---------+. (5)
                    | ` .     |  ` .
                    | (0) +---+-----+ (1)
                    |     |   |     |
                (7) +-----+---+. (6)|
                    ` .   |     ` . |
                    (3) ` +---------+ (2)
                box_corner_vertices = [
                    [xmin, ymin, zmin],
                    [xmax, ymin, zmin],
                    [xmax, ymax, zmin],
                    [xmin, ymax, zmin],
                    [xmin, ymin, zmax],
                    [xmax, ymin, zmax],
                    [xmax, ymax, zmax],
                    [xmin, ymax, zmax],
                ]
            - ``scores``: ``torch.FloatTensor`` of shape
                [num_boxes] containing detection scores for the boxes.
            - ``labels``: ``torch.IntTensor`` of shape
                [num_boxes] containing 0-indexed detection classes for the boxes.

            target: A list consisting of dictionaries each containing the key-values
            (each dictionary corresponds to a single image):
            - ``boxes``: ``torch.FloatTensor`` of shape
                [num_boxes, 8, 3] containing `num_boxes` ground truth boxes of the format
                specified in the contructor.
            - ``labels``: ``torch.IntTensor`` of shape
                [num_boxes] containing 1-indexed ground truth classes for the boxes.

        """
        # TODO: consider add input validation

        for item in preds:
            boxes = _fix_empty_tensors(item["boxes"])
            if hasattr(self, "detection_boxes"):
                self.detection_boxes.append(boxes)
            else:
                self.detections.append(boxes)

            self.detection_labels.append(item["labels"])
            self.detection_scores.append(item["scores"])

        for item in target:
            boxes = _fix_empty_tensors(item["boxes"])
            if hasattr(self, "groundtruth_boxes"):
                self.groundtruth_boxes.append(boxes)
            else:
                self.groundtruths.append(boxes)
            self.groundtruth_labels.append(item["labels"])

    def _compute_iou(self, id: int, class_id: int, max_det: int) -> torch.Tensor:
        """
        Overloading base class function.
        Computes the Intersection over Union (IoU) for ground truth and detection bounding boxes for the given
        image and class.

        Args:
            id:
                Image Id, equivalent to the index of supplied samples
            class_id:
                Class Id of the supplied ground truth and detection labels
            max_det:
                Maximum number of evaluated detection bounding boxes
        """
        if hasattr(self, "detection_boxes"):
            gt = self.groundtruth_boxes[id]
            det = self.detection_boxes[id]
        else:
            gt = self.groundtruths[id]
            det = self.detections[id]
        gt_label_mask = self.groundtruth_labels[id] == class_id
        det_label_mask = self.detection_labels[id] == class_id
        if len(gt_label_mask) == 0 or len(det_label_mask) == 0:
            return torch.Tensor([])
        gt = gt[gt_label_mask]
        det = det[det_label_mask]
        if len(gt) == 0 or len(det) == 0:
            return torch.Tensor([])

        # Sort by scores and use only max detections
        scores = self.detection_scores[id]
        scores_filtered = scores[self.detection_labels[id] == class_id]
        inds = torch.argsort(scores_filtered, descending=True)
        det = det[inds]
        if len(det) > max_det:
            det = det[:max_det]

        # generalized_box_iou
        # both det and gt are List of "boxes"
        ious = box3d_overlap_wrapper(det, gt).iou
        return ious

    def _evaluate_image(
        self,
        id: int,
        class_id: int,
        area_range: Tuple[int, int],
        max_det: int,
        ious: Dict,
    ) -> Optional[Dict]:
        """
        Overloading base class function.
        Perform evaluation for single class and image.

        Args:
            id:
                Image Id, equivalent to the index of supplied samples.
            class_id:
                Class Id of the supplied ground truth and detection labels.
            area_range:
                List of lower and upper bounding box area threshold.
            max_det:
                Maximum number of evaluated detection bounding boxes.
            ious:
                IoU results for image and class.
        """
        if hasattr(self, "detection_boxes"):
            gt = self.groundtruth_boxes[id]
            det = self.detection_boxes[id]
        else:
            gt = self.groundtruths[id]
            det = self.detections[id]

        gt_label_mask = self.groundtruth_labels[id] == class_id
        det_label_mask = self.detection_labels[id] == class_id
        if len(gt_label_mask) == 0 or len(det_label_mask) == 0:
            return None
        gt = gt[gt_label_mask]
        det = det[det_label_mask]
        if len(gt) == 0 and len(det) == 0:
            return None

        areas = box3d_volume(gt)
        ignore_area = (areas < area_range[0]) | (areas > area_range[1])

        # sort detection highest score first, sort gt ignore last
        ignore_area_sorted, gtind = torch.sort(ignore_area.to(torch.uint8))
        # Convert to uint8 temporarily and back to bool, because "Sort currently does not support bool dtype on CUDA"
        ignore_area_sorted = ignore_area_sorted.to(torch.bool)
        gt = gt[gtind]
        scores = self.detection_scores[id]
        scores_filtered = scores[det_label_mask]
        scores_sorted, dtind = torch.sort(scores_filtered, descending=True)
        det = det[dtind]
        if len(det) > max_det:
            det = det[:max_det]
        # load computed ious
        ious = (
            ious[id, class_id][:, gtind]
            if len(ious[id, class_id]) > 0
            else ious[id, class_id]
        )

        nb_iou_thrs = len(self.iou_thresholds)
        nb_gt = len(gt)
        nb_det = len(det)
        gt_matches = torch.zeros(
            (nb_iou_thrs, nb_gt), dtype=torch.bool, device=det.device
        )
        det_matches = torch.zeros(
            (nb_iou_thrs, nb_det), dtype=torch.bool, device=det.device
        )
        gt_ignore = ignore_area_sorted
        det_ignore = torch.zeros(
            (nb_iou_thrs, nb_det), dtype=torch.bool, device=det.device
        )

        if torch.numel(ious) > 0:
            for idx_iou, t in enumerate(self.iou_thresholds):
                for idx_det, _ in enumerate(det):
                    m = MeanAveragePrecision._find_best_gt_match(
                        t, gt_matches, idx_iou, gt_ignore, ious, idx_det
                    )
                    if m != -1:
                        det_ignore[idx_iou, idx_det] = gt_ignore[m]
                        det_matches[idx_iou, idx_det] = 1
                        gt_matches[idx_iou, m] = 1

        # set unmatched detections outside of area range to ignore
        det_areas = box3d_volume(det)
        det_ignore_area = (det_areas < area_range[0]) | (det_areas > area_range[1])
        ar = det_ignore_area.reshape((1, nb_det))
        det_ignore = torch.logical_or(
            det_ignore,
            torch.logical_and(
                det_matches == 0, torch.repeat_interleave(ar, nb_iou_thrs, 0)
            ),
        )
        det_matches = det_matches.cpu()
        gt_matches = gt_matches.cpu()
        scores_sorted = scores_sorted.cpu()
        gt_ignore = gt_ignore.cpu()
        det_ignore = det_ignore.cpu()

        ret = {
            "dtMatches": det_matches,
            "gtMatches": gt_matches,
            "dtScores": scores_sorted,
            "gtIgnore": gt_ignore,
            "dtIgnore": det_ignore,
        }

        if self.ret_all_prec_rec:
            self.eval_imgs.append(ret)

        return ret

    def _summarize_results(
        self, precisions: torch.Tensor, recalls: torch.Tensor
    ) -> Tuple[MAPMetricResults3D, MARMetricResults]:
        """
        Overloading base class function.
        Summarizes the precision and recall values to calculate mAP/mAR.

        Args:
            precisions:
                Precision values for different thresholds
            recalls:
                Recall values for different thresholds
        """
        results = dict(precision=precisions, recall=recalls)
        map_metrics = MAPMetricResults3D()
        last_max_det_thr = self.max_detection_thresholds[-1]
        map_metrics.map = self._summarize(results, True, max_dets=last_max_det_thr)
        if 0.25 in self.iou_thresholds:
            map_metrics.map_25 = self._summarize(
                results, True, iou_threshold=0.25, max_dets=last_max_det_thr
            )
        if 0.5 in self.iou_thresholds:
            map_metrics.map_50 = self._summarize(
                results, True, iou_threshold=0.5, max_dets=last_max_det_thr
            )

        mar_metrics = MARMetricResults()
        for max_det in self.max_detection_thresholds:
            mar_metrics[f"mar_{max_det}"] = self._summarize(
                results, False, max_dets=max_det
            )

        return map_metrics, mar_metrics

    def compute(self, sem_id_to_name_mapping: Optional[Dict[int, str]] = None) -> dict:
        metrics = MeanAveragePrecision.compute(self)
        final_results = {}

        # resemble class-based results.
        if self.class_metrics:
            seen_classes = self._get_classes()
            if sem_id_to_name_mapping is None:
                logger.warning("No sem_id to name mapping. Falling back on id=name")
                sem_id_to_name_mapping = {
                    sem_id: str(sem_id) for sem_id in seen_classes
                }

            for k, v in metrics.items():
                # Deal with per-class metrics
                if "per_class" in k:
                    # populate per class numbers
                    mapped, unmapped = set(), set()
                    for idx, pcr in enumerate(v):
                        if seen_classes[idx] not in sem_id_to_name_mapping:
                            unmapped.add(seen_classes[idx])
                        else:
                            mapped.add(seen_classes[idx])
                            final_results[
                                f"{k}@{sem_id_to_name_mapping[seen_classes[idx]]}"
                            ] = pcr
                    if len(unmapped) > 0:
                        logger.warning(
                            f"Mapped sem_ids {mapped} but DID NOT MAP sem_ids {unmapped}"
                        )
                else:
                    final_results[k] = v
        else:
            final_results = metrics
        return final_results



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import os

from typing import Dict, List, Tuple

import torch
import webdataset as wds

from atek.data_preprocess.atek_data_sample import AtekDataSample
from atek.util.file_io_utils import separate_tensors_from_dict
from atek.util.tensor_utils import concat_list_of_tensors
from omegaconf import DictConfig

SEMIDENSE_POINTS_FIELDS = [
    "msdpd#points_world",
    "msdpd#points_dist_std",
    "msdpd#points_inv_dist_std",
]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def convert_atek_sample_dict_to_wds_dict(
    index: int,
    atek_sample_dict: Dict,
    prefix_string: str,
) -> Dict:

    wds_dict = {"__key__": f"{prefix_string}_AtekDataSample_{index:06}"}

    for atek_key, atek_value in atek_sample_dict.items():
        # Semidense point data needs special handling later
        if atek_key in SEMIDENSE_POINTS_FIELDS:
            continue

        # GT needs special handling. The  tensors in the dict needs to be serialized as `.pth` file, while the rest stays in the dict.
        elif atek_key == "gt_data":
            gt_dict_no_tensor, tensor_dict = separate_tensors_from_dict(atek_value)
            wds_dict["gt_data.json"] = gt_dict_no_tensor
            for tensor_key, tensor_value in tensor_dict.items():
                wds_dict[f"gt_data#{tensor_key}.pth"] = tensor_value
            continue

        # Depth images should be directly saved as tensors
        elif atek_key.endswith("depth+images"):
            wds_dict[f"{atek_key}.pth"] = atek_value

        # Images needs to be separated into per-frame jpeg files
        elif atek_key.endswith("images"):
            assert isinstance(atek_value, torch.Tensor)
            # Transpose dimensions, [Frame, C, H, W] -> [Frame, H, W, C]
            image_frames_in_np = atek_value.numpy().transpose(0, 2, 3, 1)
            for id, img in enumerate(image_frames_in_np):
                new_key = f"{atek_key}_{id}.jpeg"
                wds_dict[new_key] = img if img.shape[-1] == 3 else img.squeeze()
            continue

        # For other fields, simply adds proper file extension to the key
        elif isinstance(atek_value, torch.Tensor):
            wds_dict[f"{atek_key}.pth"] = atek_value
        elif isinstance(atek_value, str):
            wds_dict[f"{atek_key}.txt"] = atek_value
        elif isinstance(atek_value, dict):
            wds_dict[f"{atek_key}.json"] = atek_value
        else:
            raise ValueError(f"Unsupported type {type(atek_value)} in ATEK WDS writer.")

    # Semidense point data needs to be flattended from List[Tensor (N, 3)] to a Tensor (M, 3) in order to be writable to WDS.
    # Therefore we store a `stacked_points_world` (Tensor [M, 3]) along with `points_world_lengths` (Tensor [num_frames]),
    # in order to unpack the stacked tensor later. Same for `points_inv_dist_std`.
    # obtain the "lengths" of each tensor in list.
    len_tensors = None
    for semidense_key in SEMIDENSE_POINTS_FIELDS:
        if semidense_key in atek_sample_dict:
            concatenated_tensor, current_len_tensors = concat_list_of_tensors(
                atek_sample_dict[semidense_key]
            )
            wds_dict[f"{semidense_key}+stacked.pth"] = concatenated_tensor
            if len_tensors is None:
                len_tensors = current_len_tensors.clone()
            else:
                assert torch.allclose(
                    len_tensors, current_len_tensors, atol=1
                ), f"The lengths for all semidense points data types should be the same! Instead got\n {current_len_tensors} in {semidense_key} vs \n {len_tensors}"
    if len_tensors is not None:
        wds_dict["msdpd#points_world_lengths.pth"] = len_tensors

    return wds_dict


class AtekWdsWriter:
    def __init__(self, output_path: str, conf: DictConfig) -> None:
        """
        INIT_DOC_STRING
        """
        self.output_path = output_path
        self.prefix_string = conf.prefix_string
        self.sink = None
        self.current_sample_idx = 0
        self.max_samples_per_shard = conf.max_samples_per_shard
        self.samples_in_current_shard = 0

        # A flag to indicate to remove last tar file, if it is not full. Default is False.
        self.remove_last_tar_if_not_full = (
            conf.remove_last_tar_if_not_full
            if "remove_last_tar_if_not_full" in conf
            else False
        )

    def add_sample(self, data_sample: AtekDataSample):
        """
        Add a sample to the WDS writer.
        """
        atek_sample_dict = data_sample.to_flatten_dict()
        wds_dict = convert_atek_sample_dict_to_wds_dict(
            self.current_sample_idx,
            atek_sample_dict=atek_sample_dict,
            prefix_string=self.prefix_string,
        )

        if self.sink is None:
            if not os.path.exists(self.output_path):
                os.makedirs(self.output_path)

            self.sink = wds.ShardWriter(
                f"{self.output_path}/shards-%04d.tar",
                maxcount=self.max_samples_per_shard,
            )

        self.sink.write(wds_dict)
        self.current_sample_idx += 1
        self.samples_in_current_shard += 1

    def get_num_samples(self):
        return self.current_sample_idx

    def close(self):
        """
        Close the WDS writer and flush any remaining data to disk.
        """
        if self.sink is not None:
            self.sink.close()

        # Remove the last tar file if it has less than max_samples_per_shard samples
        if self.remove_last_tar_if_not_full:
            tar_files = [f for f in os.listdir(self.output_path) if f.endswith(".tar")]
            if (
                len(tar_files) > 0
                and self.current_sample_idx % self.max_samples_per_shard != 0
            ):
                last_tar_file = os.path.join(self.output_path, tar_files[-1])
                os.remove(last_tar_file)
            else:
                logger.warning("No tar files found in the output path.")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, List

from atek.data_loaders.atek_wds_dataloader import (
    load_atek_wds_dataset,
    simple_list_collation_fn,
)

from webdataset.filters import pipelinefilter


class DepthAnything2ModelAdaptor:
    """
    A simple model adaptor class to convert ATEK WDS data to Depth Anything2 format, essentially just keep the RGB image for inference.
    Currently used for inference examples.
    """

    @staticmethod
    def get_dict_key_mapping_all():
        dict_key_mapping = {"mfcd#camera-rgb+images": "image"}
        return dict_key_mapping

    def atek_to_depth_anything2(self, data):
        for atek_wds_sample in data:
            sample = {}
            # Add images
            # from [1, C, H, W] to [H, W, C]
            image_torch = atek_wds_sample["image"].clone().detach()
            image_np = image_torch.squeeze(0).permute(1, 2, 0).numpy()
            sample["image"] = image_np
            yield sample


def load_atek_wds_dataset_as_depth_anything2(
    urls: List,
    batch_size: int,
    repeat_flag: bool,
    shuffle_flag: bool = False,
):
    adaptor = DepthAnything2ModelAdaptor()

    return load_atek_wds_dataset(
        urls,
        batch_size=batch_size,
        dict_key_mapping=DepthAnything2ModelAdaptor.get_dict_key_mapping_all(),
        data_transform_fn=pipelinefilter(adaptor.atek_to_depth_anything2)(),
        collation_fn=simple_list_collation_fn,
        repeat_flag=repeat_flag,
        shuffle_flag=shuffle_flag,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import numpy as np

import torch
from atek.data_preprocess.processors.mps_online_calib_processor import (
    MpsOnlineCalibProcessor,
)
from atek.data_preprocess.processors.mps_semidense_processor import (
    MpsSemiDenseProcessor,
)

from atek.data_preprocess.processors.mps_traj_processor import MpsTrajProcessor
from omegaconf import OmegaConf
from scipy.spatial.transform import Rotation as R

# test data paths
TEST_DIR = os.getenv("TEST_FOLDER")
CONFIG_PATH = os.getenv("CONFIG_PATH")


class MpsTrajProcessorTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_get_traj_data(self) -> None:
        conf = OmegaConf.load(CONFIG_PATH)

        mps_traj_processor = MpsTrajProcessor(
            mps_closedloop_traj_file=os.path.join(TEST_DIR, "test_mps_traj.csv"),
            conf=conf.processors.mps_traj,
        )

        # Test for query for closed loop traj data
        query_timestamp = 100_001_000  # actual timestamp is 100_000_000
        Ts_world_device, capture_timestamps, gravity_in_world = (
            mps_traj_processor.get_closed_loop_pose_by_timestamps_ns([query_timestamp])
        )
        gt_translation = torch.tensor(
            [-7.7852104154284, 0.536540244562757, 1.2436284253971863],
            dtype=torch.float32,
        )
        gt_gravity = torch.tensor([0, 0, -9.81], dtype=torch.float32)

        self.assertTrue(
            torch.allclose(
                Ts_world_device[0, :, 3].squeeze(), gt_translation, atol=1e-6
            )
        )
        self.assertTrue(
            torch.allclose(
                capture_timestamps,
                torch.tensor([100_000_000], dtype=torch.int64),
                atol=10,
            )
        )

        self.assertTrue(torch.allclose(gravity_in_world, gt_gravity))


class MpsSemiDenseProcessorTest(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_get_semidense_data(self) -> None:
        conf = OmegaConf.load(CONFIG_PATH)
        conf = conf.processors.mps_semidense
        OmegaConf.update(conf, "selected", True)
        OmegaConf.update(conf, "tolerance_ns", 10_000_000)

        mps_semidense_processor = MpsSemiDenseProcessor(
            mps_semidense_points_file=os.path.join(
                TEST_DIR, "test_mps_semidense_points.csv"
            ),
            mps_semidense_observations_file=os.path.join(
                TEST_DIR, "test_mps_semidense_observations.csv"
            ),
            conf=conf,
        )

        # test case 1: query exact timestamps
        maybe_result = mps_semidense_processor.get_semidense_points_by_timestamps_ns(
            timestamps_ns=[1_000_000_000, 2_000_000_000]
        )
        self.assertTrue(maybe_result is not None)
        points = maybe_result.points_world
        points_dist = maybe_result.points_dist_std
        points_inv_dist = maybe_result.points_inv_dist_std

        # check tensor sizes
        self.assertEqual(len(points), 2)
        self.assertEqual(points[0].shape, torch.Size([2, 3]))  # 2 points in frame 0
        self.assertEqual(points[1].shape, torch.Size([3, 3]))  # 3 points in frame 1
        points_0_gt = torch.tensor(
            [[1.1, 1.2, 1.3], [37.1, 37.2, 37.3]], dtype=torch.float32
        )
        self.assertTrue(torch.allclose(points[0], points_0_gt, atol=1e-6))

        # check dist std
        self.assertEqual(len(points_dist), 2)
        self.assertEqual(points_dist[0].shape, torch.Size([2]))  # 2 points in frame 0
        self.assertEqual(points_dist[1].shape, torch.Size([3]))  # 3 points in frame 1
        points_dist_0_gt = torch.tensor([0.002598, 0.035042], dtype=torch.float32)
        self.assertTrue(torch.allclose(points_dist[0], points_dist_0_gt, atol=1e-6))
        points_dist_1_gt = torch.tensor(
            [0.017871, 0.002598, 0.013965], dtype=torch.float32
        )
        self.assertTrue(torch.allclose(points_dist[1], points_dist_1_gt, atol=1e-6))

        # check inv dist std
        self.assertEqual(len(points_inv_dist), 2)
        self.assertEqual(
            points_inv_dist[0].shape, torch.Size([2])
        )  # 2 points in frame 0
        self.assertEqual(
            points_inv_dist[1].shape, torch.Size([3])
        )  # 3 points in frame 1
        points_inv_dist_0_gt = torch.tensor([0.003901, 0.00644], dtype=torch.float32)
        self.assertTrue(
            torch.allclose(points_inv_dist[0], points_inv_dist_0_gt, atol=1e-6)
        )
        points_inv_dist_1_gt = torch.tensor(
            [0.001829, 0.003901, 0.004357], dtype=torch.float32
        )
        self.assertTrue(
            torch.allclose(points_inv_dist[1], points_inv_dist_1_gt, atol=1e-6)
        )

        # check tracking timestamps
        gt_timestamps = torch.tensor([1_000_000_000, 2_000_000_000], dtype=torch.int64)
        self.assertTrue(
            torch.allclose(maybe_result.capture_timestamps_ns, gt_timestamps)
        )

        # check vol max and min
        self.assertTrue(isinstance(maybe_result.points_volumn_min, torch.Tensor))
        self.assertEqual(maybe_result.points_volumn_min.shape, torch.Size([3]))
        self.assertTrue(isinstance(maybe_result.points_volumn_max, torch.Tensor))
        self.assertEqual(maybe_result.points_volumn_max.shape, torch.Size([3]))


class MpsOnlineCalibProcessorTest(unittest.TestCase):
    """
    Tests for reading online calibration data from jsonl files, we need to test the MpsOnlineCalibData:
    capture_timestamps_ns=capture_timestamp_tensor,
    utc_timestamps_ns=utc_timestamp_tensor, we don't need to test this, since in the test file, this field is not valid
    projection_params=projection_params_tensor,
    ts_device_camera=ts_device_camera_tensor,
    """

    def setUp(self) -> None:
        super().setUp()

    def test_get_online_calib_data(self) -> None:
        conf = OmegaConf.load(CONFIG_PATH)
        mps_online_calib_processor = MpsOnlineCalibProcessor(
            mps_online_calib_data_file_path=os.path.join(
                TEST_DIR, "test_mps_online_calibration.jsonl"
            ),
            conf=conf.processors.mps_online_calib,
        )

        mps_online_calib_data = (
            mps_online_calib_processor.get_online_calibration_by_timestamps_ns(
                timestamps_ns=[14580434854000]
            )
        )
        self.assertTrue(mps_online_calib_data is not None)
        self.assertEqual(
            mps_online_calib_data.projection_params.shape, torch.Size([1, 3, 15])
        )
        self.assertEqual(
            mps_online_calib_data.ts_device_camera.shape, torch.Size([1, 3, 3, 4])
        )
        self.assertEqual(
            mps_online_calib_data.capture_timestamps_ns,
            torch.tensor([14580434854000], dtype=torch.int64),
        )
        self._test_translation_and_quaternion(mps_online_calib_data.ts_device_camera)
        self._test_projection_params(mps_online_calib_data.projection_params)
        return

    def _test_translation_and_quaternion(self, ts_device_camera_tensor):
        expected_translations = [
            [0.0004328977150926845, -0.00009016214024268332, -0.00004893689922337574],
            [0.006082919423532053, -0.11158645596563903, -0.0878174381882055],
            [-0.004280716004834114, -0.01184173316010912, -0.005113984064522773],
        ]

        expected_quaternions = [
            [
                0.9999746350011375,
                0.006457900925989656,
                -0.002989648005019753,
                -0.0002947452769143969,
            ],
            [
                0.7875179103510879,
                0.6156645328476295,
                0.006744113116638467,
                0.026967402693551194,
            ],
            [
                0.94426112424187,
                0.3241397857173833,
                0.0402122625965168,
                0.04107678781790648,
            ],
        ]
        for i in range(len(expected_translations)):
            expected_translation = np.array(expected_translations[i])
            expected_quaternion = np.array(expected_quaternions[i])
            # Extract the translation and rotation matrix from the tensor
            matrix = ts_device_camera_tensor[0, i, :, :].numpy()
            actual_translation = matrix[:3, 3]
            rotation_matrix = matrix[:3, :3]
            # Convert rotation matrix to quaternion
            actual_quaternion = R.from_matrix(rotation_matrix).as_quat()
            # Convert to the same format as the expected quaternion (scalar-first)
            actual_quaternion = np.roll(actual_quaternion, shift=1)
            np.testing.assert_allclose(
                actual_translation,
                expected_translation,
                atol=1e-5,
                err_msg=f"Translation mismatch at index {i}",
            )
            # Compare quaternions
            np.testing.assert_allclose(
                actual_quaternion,
                expected_quaternion,
                atol=1e-5,
                err_msg=f"Quaternion mismatch at index {i}",
            )

    def _test_projection_params(self, projection_params):
        projection_params_expect_tesnor = torch.tensor(
            [
                [
                    [
                        241.38275146484375,
                        316.2264709472656,
                        236.6782989501953,
                        -0.02765783481299877,
                        0.10371126979589462,
                        -0.07342791557312012,
                        0.012858348898589611,
                        0.0013407421065494418,
                        -0.000457167363492772,
                        0.0004445339145604521,
                        -0.0019325445173308253,
                        0.00015349453315138817,
                        0.00026054191403090954,
                        0.0033760429359972477,
                        0.00013414736895356327,
                    ],
                    [
                        241.03900146484375,
                        317.6932373046875,
                        238.18008422851562,
                        -0.02463776431977749,
                        0.09580568224191666,
                        -0.06505261361598969,
                        0.009101307950913906,
                        0.0019199964590370655,
                        -0.00044897140469402075,
                        0.0015517562860623002,
                        -0.0005500924889929593,
                        -0.000873647746630013,
                        0.00021656886383425444,
                        0.0011815830366685987,
                        0.0001968961296370253,
                    ],
                    [
                        1221.882015735315,
                        1462.729668220901,
                        1465.929424505965,
                        0.4060356696288849,
                        -0.489948419647729,
                        0.1745652818132035,
                        1.132983686620576,
                        -1.701635218233742,
                        0.6511555293441647,
                        0.0006211469747214578,
                        0.00001932200015697112,
                        -0.00001485525650871087,
                        0.0002601225712292815,
                        -0.0006582109778598294,
                        0.00003761395141407565,
                    ],
                ]
            ],
            dtype=torch.float32,
        )
        self.assertTrue(
            torch.allclose(
                projection_params,
                projection_params_expect_tesnor,
                atol=1e-6,
            )
        )
